import std.stdio;
import std.stream;
import std.string;
import std.conv;
import std.regex;
import std.file;
import Vec3D;
import Object3D;

class OBJ3D
{
  BufferedFile infile ;
  Object3D     obj ;
  Regex!(char) getfstwd ;
  string       filepath ;
  string       curmaterial ;
  Regex!(char) obj_reg_tbl[string] ;
  Regex!(char) mtl_reg_tbl[string] ;
  int delegate(string, string) obj_act_tbl[string] ;
  int delegate(string, string) mtl_act_tbl[string] ;

  this(){
    string nume = r"[-]*[0-9]+\.[0-9]+[eE]{1}[-+]?[0-9]+" ;
    string numd = r"[-]*[0-9]+\.[0-9]+" ;
    string numi = r"[-]*[0-9]+" ;
    string num  ;
    
    num = "(" ~ nume ~ "|" ~ numd ~ "|" ~ numi ~ ")" ;

    obj_reg_tbl["mtllib"] = regex( r"^mtllib\s+(\S+)"                             ,"") ;
    obj_reg_tbl["o"     ] = regex( r"^o\s+(\S+)"                                  ,"") ;
    obj_reg_tbl["s"     ] = regex( r"^s\s+(\S+)"                                  ,"") ;
    obj_reg_tbl["v"     ] = regex( r"^v\s+"  ~ num ~ r"\s+" ~ num ~ r"\s+" ~ num  ,"") ;
    obj_reg_tbl["vn"    ] = regex( r"^vn\s+" ~ num ~ r"\s+" ~ num ~ r"\s+" ~ num  ,"") ;
    obj_reg_tbl["vt"    ] = regex( r"^vt\s+" ~ num ~ r"\s+" ~ num                 ,"") ;
    obj_reg_tbl["g"     ] = regex( r"^g\s+(\S+)"                                  ,"") ;
    obj_reg_tbl["usemtl"] = regex( r"^usemtl\s+(\S+)"                             ,"") ;
    obj_reg_tbl["f"     ] = regex( r"^f\s+([0-9]*)/([0-9]*)/([0-9]*)\s+([0-9]*)/([0-9]*)/([0-9]*)\s+([0-9]*)/([0-9]*)/([0-9]*)","") ;
    obj_reg_tbl["#"     ] = regex( r"^#"                                          ,"") ;

    obj_act_tbl["mtllib"] = &read_mtllib;
    obj_act_tbl["o"     ] = &alloc_obj  ;
    obj_act_tbl["v"     ] = &add_vertex ;
    obj_act_tbl["vn"    ] = &add_normal ;
    obj_act_tbl["vt"    ] = &add_texuv  ;
    //    obj_act_tbl["g"     ] = ;
    obj_act_tbl["usemtl"] = &chg_mtl    ;
    obj_act_tbl["f"     ] = &add_face   ;

    mtl_reg_tbl["newmtl"] = regex( r"^newmtl\s+(\S+)"                             ,"") ;
    mtl_reg_tbl["Ns"    ] = regex( r"^Ns\s+" ~ num                                ,"") ;
    mtl_reg_tbl["d"     ] = regex( r"^d\s+" ~ num                                 ,"") ;
    mtl_reg_tbl["illum" ] = regex( r"^illum\s+" ~ num                             ,"") ;
    mtl_reg_tbl["Kd"    ] = regex( r"^Kd\s+" ~ num ~ r"\s+" ~ num ~ r"\s+" ~ num  ,"") ;
    mtl_reg_tbl["Ka"    ] = regex( r"^Ka\s+" ~ num ~ r"\s+" ~ num ~ r"\s+" ~ num  ,"") ;
    mtl_reg_tbl["Ks"    ] = regex( r"^Ks\s+" ~ num ~ r"\s+" ~ num ~ r"\s+" ~ num  ,"") ;
    mtl_reg_tbl["Ke"    ] = regex( r"^Ke\s+" ~ num ~ r"\s+" ~ num ~ r"\s+" ~ num  ,"") ;
    mtl_reg_tbl["map_Kd"] = regex( r"^map_Kd\s+(\S+)"                             ,"") ;
    
    mtl_act_tbl["newmtl"] = &add_mtl ;
    mtl_act_tbl["Ns"    ] = &add_mtl_Ns ;
    mtl_act_tbl["Kd"    ] = &add_mtl_Kd ;
    mtl_act_tbl["Ka"    ] = &add_mtl_Ka ;
    mtl_act_tbl["Ks"    ] = &add_mtl_Ks ;
    mtl_act_tbl["map_Kd"] = &add_mtl_mapKd ;

    curmaterial = "" ;
    
    getfstwd = regex(r"^(\S+)","") ;

  }

  ~this(){
    close() ;
    delete obj ;
  }
  
  int open(string fname, double scale=1.0){
    int i,len ;

    infile = new BufferedFile(fname) ;
    //writef("file=%s\n",fname) ;

    len=0 ;
    for( i=0 ; i<fname.length ; i++ ){
      if( fname[i]=='/' || fname[i]=='\\' ){
	len=i-1 ;
      }
    }
    if( len == 0 ){
      filepath = "" ;
    }else{
      len++ ;
      filepath = fname[0..len] ;
    }
    filepath =  std.array.replace(filepath, "\\", "/") ;

    obj = new Object3D(scale) ;
    obj.init() ;
    
    return(0) ;
  }

  int close(){
    if( !(infile is null) ){
      infile.close() ;
    }
    infile=null ;
    return(0) ;
  }

  int read() {
    while(true){
      string line=cast(string)infile.readLine() ;
      if( line is null ) line="" ;
      if( line.length==0 ){
	if( infile.eof()==true ) break ;
	continue ;
      }
      
      //      writef("%d : %s\n",i++,line) ;
      auto token = match(line, getfstwd) ;
      if( token.empty==false ){
	auto capt = token.captures ;
	string fstword = capt[1] ;
	if( (match(fstword, obj_reg_tbl["#"])).empty==false ) continue ;

	if( (fstword in obj_reg_tbl) == null ){
	  writef("keyword %s not found\n", fstword) ;
	  continue ;
	}
	if( (fstword in obj_act_tbl) != null ){
	  obj_act_tbl[fstword](fstword,line) ;
	}else{
	  //writef("internal error\n") ;
	  continue ;
	}
      }
    }

    obj.finish() ;
    return(0) ;
  }

  int read_mtllib(string act, string line){
    BufferedFile inmtl ;
    //    writef("mtllib=%s\n",line) ;
    {
      auto token = match(line, obj_reg_tbl[act]) ;
      if( token.empty==true ){
	return(0) ;
      }
      auto capt = token.captures ;
      string mtlfile = capt[1] ;
      if( filepath.length!=0 ){
	mtlfile = filepath ~ '/' ~ mtlfile ;
      }
      inmtl = new BufferedFile(mtlfile) ;
    }

    while(true){
      string mtlline=cast(string)inmtl.readLine() ;
      if( mtlline is null ) mtlline="" ;
      if( mtlline.length==0 ){
	if( inmtl.eof()==true ) break ;
	continue ;
      }
      auto token = match(mtlline, getfstwd) ;
      if( token.empty==false ){
	auto capt = token.captures ;
	string fstword = capt[1] ;
	if( (match(fstword, obj_reg_tbl["#"])).empty==false ) continue ;
      
	if( (fstword in mtl_reg_tbl) == null ){
	  writef("Keyword %s not found\n", fstword) ;
	  continue ;
	}
	if( (fstword in mtl_act_tbl) != null ){
	  mtl_act_tbl[fstword](fstword,mtlline) ;
	}else{
	  //writef("internal error\n") ;
	  continue ;
	}
      }
    }
    inmtl.close() ;

    obj.finish_material() ;
    
    //    writef("read_mtllib\n") ;
    return(0) ;
  }

  int add_mtl(string act, string line){
    auto token = match(line, mtl_reg_tbl[act]) ;
    if( token.empty==false ){
      auto capt = token.captures ;
      obj.add_material(capt[1]) ;
    }
    return(0) ;
  }

  int add_mtl_Ns(string act, string line){
    auto token = match(line,mtl_reg_tbl[act]) ;
    if( token.empty==false ){
      auto capt = token.captures ;
      obj.add_mtl_Ns( std.conv.to!(double)(capt[1]) ) ;
    }
    return(0) ;
  }

  int add_mtl_Kd(string act, string line){
    auto token = match(line,mtl_reg_tbl[act]) ;
    if( token.empty==false ){
      auto capt = token.captures ;
      double r = std.conv.to!(double)(capt[1]) ;
      double g = std.conv.to!(double)(capt[2]) ;
      double b = std.conv.to!(double)(capt[3]) ;
      obj.add_mtl_Kd(r,g,b) ;
    }
    return(0) ;
  }

  int add_mtl_Ka(string act, string line){
    auto token = match(line,mtl_reg_tbl[act]) ;
    if( token.empty==false ){
      auto capt = token.captures ;
      double r = std.conv.to!(double)(capt[1]) ;
      double g = std.conv.to!(double)(capt[2]) ;
      double b = std.conv.to!(double)(capt[3]) ;
      obj.add_mtl_Ka(r,g,b) ;
    }
    return(0) ;
  }

  int add_mtl_Ks(string act, string line){
    auto token = match(line,mtl_reg_tbl[act]) ;
    if( token.empty==false ){
      auto capt = token.captures ;
      double r = std.conv.to!(double)(capt[1]) ;
      double g = std.conv.to!(double)(capt[2]) ;
      double b = std.conv.to!(double)(capt[3]) ;
      obj.add_mtl_Ks(r,g,b) ;
    }
    return(0) ;
  }

  int add_mtl_mapKd(string act, string line){
    auto token = match(line,mtl_reg_tbl[act]) ;
    if( token.empty==false ){
      auto capt = token.captures ;
      obj.add_mtl_mapKd( filepath ~ '/' ~ capt[1] ) ;
    }
    return(0) ;
  }

  int alloc_obj(string act, string line){
    return(0) ;
  }
  
  int add_vertex(string act, string line){
    auto token = match(line,obj_reg_tbl[act]) ;
    if( token.empty==false ){
      auto capt = token.captures ;
      double x = std.conv.to!(double)(capt[1]) ;
      double y = std.conv.to!(double)(capt[2]) ;
      double z = std.conv.to!(double)(capt[3]) ;
      obj.add_vertex(x,y,z) ;
    }
    return(0) ;
  }

  int add_normal(string act, string line){
    auto token = match(line,obj_reg_tbl[act]) ;
    if( token.empty==false ){
      auto capt = token.captures ;
      double x = std.conv.to!(double)(capt[1]) ;
      double y = std.conv.to!(double)(capt[2]) ;
      double z = std.conv.to!(double)(capt[3]) ;
      obj.add_normal(x,y,z) ;
    }
    return(0) ;
  }

  int add_texuv(string act, string line){
    auto token = match(line,obj_reg_tbl[act]) ;
    if( token.empty==false ){
      auto capt = token.captures ;
      double u = std.conv.to!(double)(capt[1]) ;
      double v = std.conv.to!(double)(capt[2]) ;
      obj.add_texuv(u,v) ;
    }
    return(0) ;
  }

  int chg_mtl(string act, string line){
    auto token = match(line,obj_reg_tbl[act]) ;
    if( token.empty==false ){
      auto capt = token.captures ;
      curmaterial = capt[1] ~ "" ;
    }
    return(0) ;
  }

  int add_face(string act, string line){
    face3D   f,t,n ;
    
    auto token = match(line,obj_reg_tbl[act]) ;
    if( token.empty==false ){
      auto capt = token.captures ;
      if( capt[1].length!=0 && capt[4].length!=0 && capt[7].length!=0 ){
	f.p0=std.conv.to!(int)(capt[1]) ;
	f.p1=std.conv.to!(int)(capt[4]) ;
	f.p2=std.conv.to!(int)(capt[7]) ;
      }
      if( capt[2].length!=0 && capt[5].length!=0 && capt[8].length!=0 ){
	t.p0=std.conv.to!(int)(capt[2]) ;
	t.p1=std.conv.to!(int)(capt[5]) ;
	t.p2=std.conv.to!(int)(capt[8]) ;
      }
      if( capt[3].length!=0 && capt[6].length!=0 && capt[9].length!=0 ){
	n.p0=std.conv.to!(int)(capt[3]) ;
	n.p1=std.conv.to!(int)(capt[6]) ;
	n.p2=std.conv.to!(int)(capt[9]) ;
      }
      obj.add_face( curmaterial, f,t,n ) ;
    }
    return(0) ;
  }
  
}

