import win32.windows;
import win32frame.glut;
import std.stdio;
import core.thread;
import std.math;

import mt19937ar ;
import GameFrame;
import Sprite;
import SPformat;

import Vec3D;
import Quaternion;
import RubiksCube;

struct History{
  int type ;
  int num  ;
  int rot  ;
} ;

class myGameFrame : GameFrame
{
  this(string name, string ClassName) {
    super(name, ClassName) ;
  }

  int DestroyProc(){
    return(0) ;
  }
  
  int GameLoop(){
    int  tick =0 ;
    int  fps ;
    Sprite    sp = cast(Sprite)(SP) ;
    GameInfo* gi = cast(GameInfo*)(&GmInfo) ;
    
    init_genrand(7777777) ;

    sp.init(512, 8192, 32, 8192, gi) ;
    sp.add3DPCG("3d_obj/block.obj"              , 0, 0.1) ;

    int fonttop=32 ;
    sp.add2DPCG("img_data/font8x16.png",fonttop,  8,16,16,16,true) ;
    SPformat spformat = new SPformat(sp,fonttop,8,16) ;

    sp.setLight(0, 1.0, 1.0, 4.0, 0.0,  1.0, 1.0, 1.0,  0.3, 0.3, 0.3,  1.0, 1.0, 1.0) ;
    sp.setLight(1,-1.0, 1.0, 4.0, 0.0,  1.0, 1.0, 1.0,  0.3, 0.3, 0.3,  1.0, 1.0, 1.0) ;
    sp.setClearColor( 0.3, 0.3, 0.3 ) ;
    
    History[] his ; his.length=100 ;
    int hisptr=0 ;
    RubiksCube rb = new RubiksCube() ;
    int maxsize=6 ;
    shuffle(rb,maxsize,his) ;

    double rd_x=0 ;
    double rd_y=0 ;
    Quaternion g_rot ;
    Quaternion v_rot ;
    Quaternion rot_x ; rot_x.setRotate(rd_x,1.0, 0.0, 0.0) ;
    Quaternion rot_y ; rot_y.setRotate(rd_y,0.0, 1.0, 0.0) ;
    int speed=20 ;
    int solved_wait=0 ;
    
    while(running==true){
      double rad_rd = -cast(double)(rd_y)*PI*2.0/360.0 ;
      rot_y.setRotate(rad_rd,0.0, 1.0, 0.0) ;
      rd_y = rd_y+0.1 ;
      if( rd_y>=360.0 ) rd_y-=360.0 ;
      
      v_rot.setRotate( 3.14/4.0, 1.0, 1.0, 0.0 ) ;
      v_rot.mul(&rot_x) ;
      v_rot.mul(&rot_y) ;
      
      if( sp.resetSync()==0 ){
	Anim[] anim ;
	if( hisptr==his.length ){
	  anim = rb.makeAnimation( his[hisptr-1].type, his[hisptr-1].num, his[hisptr-1].rot,
				   0.2, 0.0) ;
	}else{
	  anim = rb.makeAnimation( his[hisptr].type, his[hisptr].num, his[hisptr].rot,
				   0.2, cast(double)(tick%speed)/cast(double)(speed)) ;
	}
	
	for( int i=0 ; i<anim.length ; i++ ){
	  g_rot.set(&v_rot) ;
	  anim[i].pos.rotate(g_rot.getAngle(),g_rot.getAxis()) ;
	  g_rot.mul(&anim[i].rot) ;
	  vec3D rotvec = g_rot.getAxis() ;
	  
	  sp.set3Draw(0,
		      anim[i].pos.x, anim[i].pos.y, anim[i].pos.z,
		      rotvec.x, rotvec.y, rotvec.z,
		      g_rot.getAngle()*360.0/(PI*2.0),
		      1.0, 1.0, 1.0, BLEND_TYPE.normal) ;
	}
	if( tick%speed==(speed-1) ){
	  if( hisptr<his.length ){
	    rb.rotate(his[hisptr].type, his[hisptr].num, his[hisptr].rot) ;
	    hisptr++ ;
	  }else{
	    if( solved_wait<15 ){
	      solved_wait++ ;
	    }else{
	      solved_wait=0 ;
	      shuffle(rb,maxsize,his) ;
	      hisptr=0 ;
	    }
	  }
	}
	
	fps=sp.getFPS() ;
	spformat.print(0+8,0+8,"fps=%d tick=%d",fps,tick) ;
	spformat.print(0+8,0+8+16,"OpenGL Version %s",getglversion()) ;
	spformat.print(0+8,0+8+32,"%s",getglrenderer()) ;
	spformat.print(0+8,0+8+48,"glext %s",getglextmsg()) ;
	tick++ ;
      }
      sp.waitSync() ;
      
    }
    return(0) ;
  }

  int shuffle( RubiksCube rb, int maxsize, History[] his ){
    rb.init(genrand_int31()%(maxsize-1)+2) ;
    for( int i=0 ; i<his.length ; i++ ){
      int type=genrand_int31()%3 ;
      int num =genrand_int31()%rb.getNum() ;
      int rot =genrand_int31()%3+1 ;
      rb.rotate(type, num, rot) ;
      his[his.length-1-i].type=type ;
      his[his.length-1-i].num =num  ;
      his[his.length-1-i].rot =4-rot ;
    }
    return(0) ;
  }
  
}
